//
// Created by pcg1 on 12.01.18.
//

#ifndef FIELDOPT_DFO_H
#define FIELDOPT_DFO_H

#include "Optimization/optimizer.h"
#include "Subproblem.h"
#include "DFO_Model.h"
#include <ncurses.h>
#include <fstream>
#include <iostream>

namespace Optimization {
namespace Optimizers {

/*!
 * @brief This is a fantastic description of the DFO method, with references.
 */
class DFO : public Optimizer {
 public:
  DFO(Settings::Optimizer *settings,
      Case *base_case,
      Model::Properties::VariablePropertyContainer *variables,
      Reservoir::Grid::Grid *grid,
      Logger *logger);

  QString GetStatusStringHeader() const {};
  QString GetStatusString() const {};
  Model::Properties::VariablePropertyContainer *variables_;
  void set_next_step(int a);

  Eigen::VectorXd ScaleVariablesFromAlgorithmToApplication(Eigen::VectorXd point);
  Eigen::VectorXd ScaleVariablesFromApplicationToAlgorithm(Eigen::VectorXd point);
  void ConvertPointToCase(Eigen::VectorXd point, Optimization::Case *new_case);

  void CreateScalingVector();

 private:

  QList<QUuid> UUIDS_in_same_order_as_in_realvar_base_case_;
  bool terminated = false;
  Eigen::VectorXd realvectest_;
  QUuid mapIndexToCase;
  vector<QUuid> mapIndicesToCase;
  Eigen::VectorXd scaling_;

  std::string filenamePoint;
  std::string filenameType;
  std::string filenameTrr;

  std::string color_from = "31";
  std::string color_to = "33";
  Model::Properties::VariablePropertyContainer *varcont_;
  DFO_Model DFO_model_;
  void iterate() override;
  int number_of_interpolation_points_;
  int number_of_variables_;
  Optimization::Case *base_case_;
  int last_action_;
  int next_step_;
  std::string get_action_name(int a);
  Eigen::VectorXd weights_distance_from_optimum_lsq_;

  int number_of_points_first_set;
  bool multiple_new_points;
  double trust_region_radius_tilde;
  double trust_region_radius_icb;
  double rho;
  bool is_model_CFL;
  Eigen::VectorXd tmp_eval;

  int number_of_crit_step_finished_with_bad_poisedness;
  int number_of_crit_step_finished_without_checking_poisedness;
  int ng;
  double alpha;
  int number_of_tiny_improvements;
  double r;
  double w;
  double u;
  double beta;
  double epsilon_c;
  double tau;
  double eta1;
  double gamma;
  double gamma_inc;
  double trust_region_radius_inc;
  double trust_region_radius_max;
  double trust_region_radius_end;

  bool notConverged;

  int crit_steps;
  int accept_steps;
  int model_impr_steps;

  Eigen::VectorXd *refFuncVals;
  Eigen::MatrixXd *refY;
  Eigen::MatrixXd *refDerivatives;

  int number_of_new_points;
  int number_of_function_calls;
  int number_of_parallell_function_calls;
  Eigen::MatrixXd new_gradients;
  Eigen::MatrixXd new_points;
  Eigen::VectorXd new_point;
  Eigen::VectorXd last_trial_point;
  Eigen::VectorXd new_gradient;
  Eigen::VectorXi new_points_indicies;
  Eigen::VectorXd function_evaluations;
  double function_evaluation;
  int new_point_index;
  int criticality_step_iteration;

  bool force_criticality_step;

  bool is_successful_iteration() {};

  void handleEvaluatedCase(Case *c);

  TerminationCondition IsFinished() {
    cout << "JUST CALLED ISFINISHED" << endl;
    return TerminationCondition::NOT_FINISHED;
  };

 protected:

 private:
  int previous_iterate_type_;
  double initial_trust_region_radius_;
  double required_poisedness_;
  QList<QUuid> realvar_uuid_;
  Settings::Optimizer *settings_;
  int iterations_;

};
}
}

#endif //FIELDOPT_DFO_H
