//
// Created by joakim on 12.06.18.
//

#ifndef FIELDOPT_VIRTUALSIMULATOR_H
#define FIELDOPT_VIRTUALSIMULATOR_H
#include <iostream>
#include <iomanip>
#include <casadi/casadi.hpp>
#include <Eigen/Core>

class VirtualSimulator {
 private:
  std::string problem;
  casadi::NlpBuilder nl;
  std::vector<casadi::DM> input;
  casadi::Function f;
  casadi::Function fj;
  casadi::Function g;
  casadi::Function gj;
  int m_; //Number of constraints (linear and nonlinear);
  int n_; //Number of variables;
  int mb_; //Number of bounds.

 public:
  VirtualSimulator(std::string problemFile);
  VirtualSimulator();
  Eigen::VectorXd evaluateConstraints(Eigen::VectorXd point);
  Eigen::MatrixXd evaluateConstraintGradients(Eigen::VectorXd point);
  double evaluateFunction(Eigen::VectorXd point);
  Eigen::VectorXd evaluateFunctionGradients(Eigen::VectorXd point);
  casadi::DM convertEigenMatrixToCasadi(Eigen::MatrixXd point);
  Eigen::MatrixXd convertCasadiMatrixToEigen(casadi::DM casadiMatrix);
  int GetNumberOfVariables();
  int GetNumberOfConstraints();
  Eigen::VectorXd GetInitialPoint();
  Eigen::VectorXd GetLowerBoundsForVariables();
  Eigen::VectorXd GetUpperBoundsForVariables();
  Eigen::VectorXd GetLowerBoundsForConstraints();
  Eigen::VectorXd GetUpperBoundsForConstraints();
  bool IsFeasiblePoint(Eigen::VectorXd point);
  Eigen::VectorXd Solve();
};

#endif //FIELDOPT_VIRTUALSIMULATOR_H
